# cython: language_level=3
import cython

cdef unsigned int OBJECT_HEADER_SIZE = 12
cdef unsigned int EXTENSION_SIZE = 4

cdef unsigned int TYPE_NUM_BITS = 16
cdef unsigned int TYPE_POS = 0

cdef unsigned int USER_NUM_BITS = 1
cdef unsigned int USER_POS = 16

cdef unsigned int EXTENDED_NUM_BITS = 1
cdef unsigned int EXTENDED_POS = 17

cdef unsigned int VERSION_NUM_BITS = 12
cdef unsigned int VERSION_POS = 20

cdef unsigned int ONLY_SUBOBJECTS_NUM_BITS = 1
cdef unsigned int ONLY_SUBOBJECTS_POS = 30

cdef unsigned int LENGTH_NUM_BITS = 30
cdef unsigned int LENGTH_POS = 0

cdef unsigned int EXTENSION_NUM_BITS = 12
cdef unsigned int EXTENSION_POS = 0

cdef class ObjectHeader:
    cdef public str endianness
    cdef readonly unsigned int type
    cdef readonly unsigned int version
    cdef readonly bint user
    cdef readonly bint extended
    cdef readonly bint only_subobjects
    cdef public unsigned long length
    cdef readonly unsigned long id
    cdef public unsigned long address

    def __repr__(self):
        return (
            'Header[{}]('.format(self.type)
            + 'endianness={}, '.format(self.endianness)
            + 'version={}, '.format(self.version)
            + 'id={}, '.format(self.id)
            + 'user={}, '.format(self.user)
            + 'extended={}, '.format(self.extended)
            + 'only_subobjects={}, '.format(self.only_subobjects)
            + 'length={}, '.format(self.length)
            + 'address={})'.format(self.address)
        )


cpdef bint bool_bit_from_pos(unsigned int uint32_word, unsigned int pos):
    '''parse a Python Boolean from a bit a position `pos` in an
    unsigned 32bit integer.
    '''
    return uint32_word & (1 << pos)


cpdef unsigned int get_bits_from_word(
    unsigned int uint32_word,
    unsigned int num_bits,
    unsigned int first,
):
    '''return `num_bits` bits from the input word
    starting at first

    assume the input word was:
        MSB                                    LSB
        0000_0000__0000_0000__1010_1100__0000_0000

    and first=10 and num_bits=4

    the return value would be: 1011 (with leading zeros)
    '''
    return (uint32_word >> first) & ((1 << num_bits) - 1)


cdef (unsigned int, unsigned int, bint, bint) parse_type_field(unsigned int word):
    '''parse TypeInfo
    '''
    type_ = get_bits_from_word(word, TYPE_NUM_BITS, TYPE_POS)
    user_bit = bool_bit_from_pos(word, USER_POS)
    extended = bool_bit_from_pos(word, EXTENDED_POS)
    version = get_bits_from_word(word, VERSION_NUM_BITS, VERSION_POS)
    return type_, version, user_bit, extended


@cython.boundscheck(False)
@cython.wraparound(False)
cdef unsigned long unpack_uint32(const unsigned char[:] data):
    return (
        (<unsigned long> data[0])
        + ((<unsigned long> data[1]) << 8)
        + ((<unsigned long> data[2]) << 16)
        + ((<unsigned long> data[3]) << 24)
    )


cpdef ObjectHeader parse_header_bytes(const unsigned char[:] header_bytes):
    cdef unsigned long type_int
    cdef unsigned long id_field
    cdef unsigned long length_field

    cdef unsigned long type_
    cdef bint user
    cdef bint extended
    cdef unsigned long version
    cdef bint only_subobjects
    cdef unsigned long length


    type_int = unpack_uint32(header_bytes[0:4])
    type_, version, user, extended = parse_type_field(type_int)
    id_field = unpack_uint32(header_bytes[4:8])

    length_field = unpack_uint32(header_bytes[8:12])

    only_subobjects = bool_bit_from_pos(length_field, ONLY_SUBOBJECTS_POS)
    length = get_bits_from_word(length_field, LENGTH_NUM_BITS, LENGTH_POS)

    header = ObjectHeader()
    header.type = type_
    header.user = user
    header.extended = extended
    header.version = version
    header.id = id_field
    header.only_subobjects = only_subobjects
    header.length = length

    return header



# cython: language_level=3
import cython
from cpython cimport array
import numpy as np
cimport numpy as np

INT16 = np.int16
ctypedef np.int16_t INT16_t


@cython.wraparound(False)  # disable negative indexing
cpdef read_sector_information(
    const unsigned char[:] data,
    unsigned long n_pixels,
    unsigned long offset = 0,
):
    cdef unsigned long i
    cdef short j, n
    cdef INT16_t* short_ptr

    cdef list sectors = []
    cdef array.array sector

    cdef unsigned long pos = 0
    for i in range(n_pixels):

        short_ptr = <INT16_t*> &data[pos + offset]
        n = short_ptr[0]
        pos += 2

        sector = array.array('h')
        for j in range(n):
            short_ptr = <INT16_t*> &data[pos + offset]
            sector.append(short_ptr[0])
            pos += 2

        # FIXME:
        # according to a comment in the c-sources
        # there is might be an old bug here,
        # which is trailing zeros.
        # is an ascending list of numbes, so any zero
        # after the first position indicates the end of sector.
        #
        # DN: maybe this bug was fixed long ago,
        # so maybe we do not have to account for it here
        # I will check for it in the tests.
        sectors.append(sector)

    return sectors, pos



@cython.wraparound(False)  # disable negative indexing
cpdef dict parse_mc_event(
    const unsigned char[:] data,
    unsigned int version
):

    cdef unsigned long pos = 0
    cdef float xcore, ycore, aweight
    cdef int shower_num

    shower_num = (<int*> &data[pos])[0]
    pos += 4
    xcore = (<float*> &data[pos])[0]
    pos += 4
    ycore = (<float*> &data[pos])[0]
    pos += 4

    if version >= 2:
        aweight = (<float*> &data[pos])[0]
    else:
        aweight = 1.0

    return {
        'shower_num': shower_num,
        'xcore': xcore,
        'ycore': ycore,
        'aweight': aweight
    }



# cython: language_level=3
import cython
import numpy as np
cimport numpy as np

INT16 = np.int16
ctypedef np.int16_t INT16_t

UINT32 = np.uint32
ctypedef np.uint32_t UINT32_t

INT32 = np.int32
ctypedef np.int32_t INT32_t

UINT64 = np.uint64
ctypedef np.uint64_t UINT64_t

INT64 = np.int64
ctypedef np.int64_t INT64_t

cdef F32 = np.float32


@cython.wraparound(False)  # disable negative indexing
cpdef (UINT64_t, unsigned int) unsigned_varint(const unsigned char[:] data, unsigned long offset=0):
    cdef unsigned int length
    cdef unsigned long value
    length = get_length_of_varint(data[offset])
    value = parse_varint(data[offset:offset + length])
    return value, length


@cython.wraparound(False)  # disable negative indexing
cpdef (INT64_t, unsigned int) varint(const unsigned char[:] data, unsigned long offset=0):
    cdef unsigned int length
    cdef unsigned long value
    value, length = unsigned_varint(data, offset)
    if (value & 1) == 1:  # Negative number;
        return -(value >> 1) - 1, length
    return value >> 1, length


cpdef unsigned int get_length_of_varint(const unsigned char first_byte):
    if (first_byte & 0x80) == 0:
        return 1
    if (first_byte & 0xc0) == 0x80:
        return 2
    if (first_byte & 0xe0) == 0xc0:
        return 3
    if (first_byte & 0xf0) == 0xe0:
        return 4
    if (first_byte & 0xf8) == 0xf0:
        return 5
    if (first_byte & 0xfc) == 0xf8:
        return 6
    if (first_byte & 0xfe) == 0xfc:
        return 7
    if (first_byte & 0xff) == 0xfe:
        return 8
    return 9


@cython.boundscheck(False)
@cython.wraparound(False)  # disable negative indexing
cpdef UINT64_t parse_varint(const unsigned char[:] var_int_bytes):
    length = var_int_bytes.shape[0]
    cdef UINT64_t v[9]
    cdef int i  = 0
    for i in range(length):
        v[i] = var_int_bytes[i]

    if length == 1:
        return v[0]

    if length == 2:
        return ((v[0] & 0x3f) <<8) | v[1]

    if length == 3:
        return (
            ((v[0] & 0x1f) << 16)
            | (v[1] << 8)
            | v[2]
        )

    if length == 4:
        return (
            ((v[0] & 0x0f) << 24)
            | (v[1] << 16)
            | (v[2] << 8)
            | v[3]
        )
    if length == 5:
        return (
            ((v[0] & 0x07) << 32)
            | (v[1] << 24)
            | (v[2] << 16)
            | (v[3] << 8)
            | v[4]
        )
    if length == 6:
        return (
            ((v[0] & 0x03) << 40)
            | (v[1] << 32)
            | (v[2] << 24)
            | (v[3] << 16)
            | (v[4] << 8)
            | v[5]
        )
    if length == 7:
        return (
            ((v[0] & 0x01) << 48)
            | (v[1] << 40)
            | (v[2] << 32)
            | (v[3] << 24)
            | (v[4] << 16)
            | (v[5] << 8)
            | v[6]
        )
    if length == 8:
        return (
            (v[1]<<48)
            | (v[2]<<40)
            | (v[3]<<32)
            | (v[4]<<24)
            | (v[5]<<16)
            | (v[6]<<8)
            | v[7]
        )

    return (
        (v[1]<<56)
        | (v[2]<<48)
        | (v[3]<<40)
        | (v[4]<<32)
        | (v[5]<<24)
        | (v[6]<<16)
        | (v[7]<<8)
        | v[8]
    )


@cython.wraparound(False)  # disable negative indexing
cpdef unsigned_varint_array(
    const unsigned char[:] data,
    unsigned long n_elements,
    unsigned long offset = 0,
):
    cdef np.ndarray[UINT64_t, ndim=1] output = np.empty(n_elements, dtype=UINT64)

    cdef unsigned long i
    cdef unsigned long pos
    cdef unsigned long idx
    cdef unsigned long length
    pos = 0

    for i in range(n_elements):
        idx = pos + offset
        length = get_length_of_varint(data[idx])
        output[i] = parse_varint(data[idx:idx + length])
        pos += length

    return output, pos


@cython.wraparound(False)  # disable negative indexing
cpdef varint_array(
    const unsigned char[:] data,
    unsigned long n_elements,
    unsigned long offset = 0,
):
    cdef unsigned long bytes_read
    cdef np.ndarray[INT64_t, ndim=1] output = np.empty(n_elements, dtype=INT64)

    cdef UINT64_t val
    cdef unsigned int length

    cdef unsigned long one = 1;
    cdef unsigned long i
    cdef unsigned long pos = 0

    for i in range(n_elements):
        length = get_length_of_varint(data[pos])
        val = parse_varint(data[pos:pos + length])
        pos += length

        if (val & one):
            output[i] = -(val >> one) - one
        else:
            output[i] = val >> one

    return output, pos


@cython.wraparound(False)  # disable negative indexing
cpdef unsigned_varint_arrays_differential(
    const unsigned char[:] data,
    unsigned long n_arrays,
    unsigned long n_elements,
    unsigned long offset = 0,
):
    cdef unsigned long pos = 0
    cdef unsigned long bytes_read = 0
    cdef unsigned long bytes_read_total = 0
    cdef unsigned long i
    cdef unsigned long j
    cdef (unsigned long, unsigned long) shape = (n_arrays, n_elements)

    cdef np.ndarray[UINT32_t, ndim=2] output = np.zeros(shape, dtype=UINT32)
    cdef UINT32_t[:, :] output_view = output
    cdef UINT32_t[:] output_view_1d

    for i in range(n_arrays):

        output_view_1d = output_view[i]

        bytes_read = unsigned_varint_array_differential(
            data, output=output_view_1d, offset=offset
        )
        offset += bytes_read
        bytes_read_total += bytes_read

    return output, bytes_read_total


@cython.wraparound(False)  # disable negative indexing
cdef unsigned long unsigned_varint_array_differential(
    const unsigned char[:] data,
    UINT32_t[:] output,
    unsigned long offset = 0,
) except -1:

    cdef unsigned long n_elements = output.shape[0]
    cdef int val = 0
    cdef unsigned long i
    cdef unsigned long pos = 0
    cdef unsigned char v0, v1, v2, v3, v4

    for i in range(n_elements):
        v0 = data[pos + offset]
        pos += 1

        if (v0 & 0x80) == 0:  # one byte
            if (v0 & 0x01) == 0:  # positive
                val += v0 >> 1
            else:  # negative
                val -= (v0 >> 1) + 1
        elif (v0 & 0xc0) == 0x80:  # two bytes
            v1 = data[pos + offset]
            pos += 1
            if (v1 & 0x01) == 0:  # positive
                val += ((v0 & 0x3f) << 7) | (v1 >> 1)
            else:  # negative
                val -= (((v0 & 0x3f) << 7) | (v1 >> 1)) + 1
        elif (v0 & 0xe0) == 0xc0:  # three bytes
            v1 = data[pos + offset + 0]
            v2 = data[pos + offset + 1]
            pos += 2

            if (v2 & 0x01) == 0:
                val += (
                    ((v0 & 0x1f) << 15)
                    | (v1 << 7)
                    | (v2 >> 1)
                )
            else:
                val -= (
                    ((v0 & 0x1f) << 15)
                    | (v1 << 7)
                    | (v2 >> 1)
                ) + 1
        elif (v0 & 0xf0) == 0xe0:  # four bytes
            v1 = data[pos + offset + 0]
            v2 = data[pos + offset + 1]
            v3 = data[pos + offset + 2]
            pos += 3
            if (v3 & 0x01) == 0:
                val += (
                    ((v0 & 0x0f) << 23)
                    | (v1 << 15)
                    | (v2 << 7)
                    | (v3 >> 1)
                )
            else:
                val -= (
                    ((v0 & 0x0f) << 23)
                    | (v1 << 15)
                    | (v2 << 7)
                    | (v3 >> 1)
                ) + 1
        elif (v0 & 0xf8) == 0xf0:
            v1 = data[pos + offset + 0]
            v2 = data[pos + offset + 1]
            v3 = data[pos + offset + 2]
            v4 = data[pos + offset + 3]
            pos += 4
            # The format would allow bits 32 and 33 being set but we ignore this here. */
            if (v4 & 0x01) == 0:
                val += (
                    ((v0 & 0x07) << 31)
                    | (v1 << 23)
                    | (v2 << 15)
                    | (v3 << 7)
                    | (v4 >> 1)
                )
            else:
                val -= (
                    ((v0 & 0x07) << 31)
                    | (v1 << 23)
                    | (v2 << 15)
                    | (v3 << 7)
                    | (v4 >> 1)
                ) + 1
        output[i] = val

    return pos



def simtel_pixel_timing_parse_list_type_2(
    const unsigned char[:] data,
    const INT16_t[:, :] pixel_list,
    int num_gains,
    int num_pixels,
    int num_types,
    bint with_sum,
    bint glob_only_selected,
    float granularity,
):
    cdef int start, stop, list_index
    cdef int i_pix, i_type
    cdef unsigned long pos = 0
    cdef unsigned int length = 0

    cdef np.ndarray[float, ndim=2] timval = np.zeros((num_pixels, num_types), dtype=np.float32)
    # The first timing element is always initialised to indicate unknown.
    timval[:, 0] = -1

    cdef np.ndarray[INT32_t, ndim=2] pulse_sum_loc = np.zeros((num_gains, num_pixels), dtype=INT32)
    cdef np.ndarray[INT32_t, ndim=2] pulse_sum_glob = np.zeros((num_gains, num_pixels), dtype=INT32)

    for list_index in range(pixel_list.shape[0]):
        start = pixel_list[list_index][0]
        stop = pixel_list[list_index][1]
        for i_pix in range(start, stop + 1):
            for i_type in range(num_types):
                timval[i_pix, i_type] = granularity # * read_short(self)
                pos += 2

            if with_sum:
                for i_gain in range(num_gains):
                    pulse_sum_loc[i_gain, i_pix], length = varint(data, offset=pos)
                    pos += length

                if glob_only_selected:
                    for i_gain in range(num_gains):
                        pulse_sum_glob[i_gain, i_pix], length = varint(data, offset=pos)
                        pos += length

    if with_sum and pixel_list.shape[0] > 0 and not glob_only_selected:
        for i_gain in range(num_gains):
            for i_pix in range(num_pixels):
                pulse_sum_glob[i_gain, i_pix], length = varint(data, offset=pos)
                pos += length

    return {
        'timval': timval,
        'pulse_sum_glob': pulse_sum_glob,
        'pulse_sum_loc': pulse_sum_loc,
    }, pos


def parse_1208(
    const unsigned char[:] data,
    int n_pixels,
    int nonempty,
    int version,
    unsigned int flags
):

    cdef unsigned int length
    cdef INT32_t* int_ptr
    cdef INT16_t* short_ptr
    cdef float* float_ptr

    cdef np.ndarray[float, ndim=1] photoelectrons = np.zeros(n_pixels, dtype=F32)
    cdef np.ndarray[INT32_t, ndim=1] photon_counts = None
    cdef list time = [[] for _ in range(n_pixels)]
    cdef list amplitude

    if flags & 1:
        amplitude = [[] for _ in range(n_pixels)]
    else:
        amplitude = None

    cdef unsigned long pos = 0
    cdef int i, j

    for i in range(nonempty):
        if version > 2:
            pix_id, length = varint(data, offset=pos)
        else:
            short_ptr = <INT16_t*> &data[pos]
            pix_id = short_ptr[0]
            length = 2

        pos += length

        int_ptr = <INT32_t*> &data[pos]
        n_pe = int_ptr[0]
        pos += 4

        photoelectrons[pix_id] = n_pe
        for j in range(n_pe):
            float_ptr = <float*> &data[pos]
            time[pix_id].append(float_ptr[0])
            pos += 4

        if flags & 1:
            for j in range(n_pe):
                float_ptr = <float*> &data[pos]
                amplitude[pix_id].append(float_ptr[0])
                pos += 4

    if flags & 4:
        photon_counts = np.zeros(n_pixels, dtype=INT32)

        int_ptr = <INT32_t*> &data[pos]
        nonempty = int_ptr[0]
        pos += 4

        for i in range(nonempty):
            short_ptr = <INT16_t*> &data[pos]
            pix_id = short_ptr[0]
            pos += 2

            int_ptr = <INT32_t*> &data[pos]
            photon_counts[pix_id] = int_ptr[0]
            pos += 4

    return photoelectrons, time, amplitude, photon_counts


cpdef simtel_pixel_timing_parse_list_type_1(
    const unsigned char[:] data,
    const INT16_t[:] pixel_list,
    int num_gains,
    int num_pixels,
    int num_types,
    bint with_sum,
    bint glob_only_selected,
    float granularity,
):
    cdef int start, stop, list_index
    cdef long pixel_list_length = pixel_list.shape[0]
    cdef int i, i_gain, i_type, i_pix
    cdef unsigned int length = 0
    cdef np.ndarray[INT64_t, ndim=1] result

    cdef np.ndarray[float, ndim=2] timval = np.zeros((num_pixels, num_types), dtype=np.float32)
    # The first timing element is always initialised to indicate unknown.
    for i in range(num_pixels):
        timval[i, 0] = -1

    cdef np.ndarray[INT32_t, ndim=2] pulse_sum_loc = np.zeros((num_gains, num_pixels), dtype=INT32)
    cdef np.ndarray[INT32_t, ndim=2] pulse_sum_glob = np.zeros((num_gains, num_pixels), dtype=INT32)

    cdef short* short_ptr

    cdef unsigned long pos = 0
    for i in range(pixel_list_length):
        i_pix = pixel_list[i]
        for i_type in range(num_types):
            short_ptr = <short*> &(data[pos])
            timval[i_pix, i_type] = granularity * (short_ptr[0])
            pos += 2

        if with_sum:
            for i_gain in range(num_gains):
                pulse_sum_loc[i_gain, i_pix], length = varint(data, offset=pos)
                pos += length

            if glob_only_selected:
                for i_gain in range(num_gains):
                    pulse_sum_glob[i_gain, i_pix], length = varint(data, offset=pos)
                    pos += length

    if with_sum and len(pixel_list) > 0 and not glob_only_selected:
        for i_gain in range(num_gains):
            for i_pix in range(num_pixels):
                pulse_sum_glob[i_gain, i_pix], length = varint(
                    data, offset=pos,
                )
                pos += length

    return {
        'timval': timval,
        'pulse_sum_glob': pulse_sum_glob,
        'pulse_sum_loc': pulse_sum_loc,
    }, pos



'''
Implementation of an EventIOFile that
loops through SimTel Array events.
'''
import re
from copy import copy
from collections import defaultdict
import logging
import warnings
from ..base import EventIOFile
from ..exceptions import check_type
from .. import iact
from ..histograms import Histograms
from .objects import (
    ADCSamples,
    ADCSums,
    ArrayEvent,
    CameraMonitoring,
    CameraOrganization,
    CameraSettings,
    CameraSoftwareSettings,
    TriggerInformation,
    DisabledPixels,
    DriveSettings,
    History,
    ImageParameters,
    LaserCalibration,
    MCEvent,
    MCPhotoelectronSum,
    MCRunHeader,
    MCShower,
    PixelList,
    PixelSettings,
    PixelTiming,
    PointingCorrection,
    RunHeader,
    TelescopeEvent,
    TelescopeEventHeader,
    TrackingPosition,
)


log = logging.getLogger(__name__)


camel_re1 = re.compile('(.)([A-Z][a-z]+)')
camel_re2 = re.compile('([a-z0-9])([A-Z])')


def camel_to_snake(name):
    s1 = camel_re1.sub(r'\1_\2', name)
    return camel_re2.sub(r'\1_\2', s1).lower()


class NoTrackingPositions(Exception):
    pass


class SimTelFile(EventIOFile):
    def __init__(self, path, allowed_telescopes=None):
        super().__init__(path)

        self.path = path
        self.allowed_telescopes = allowed_telescopes
        self.histograms = None

        self.history = []
        o = next(self)
        while isinstance(o, History):
            self.history.append(o)
            o = next(self)

        check_type(o, RunHeader)
        self.header = o.parse()
        self.n_telescopes = self.header['n_telescopes']

        o = next(self)
        self.mc_run_headers = []
        while isinstance(o, MCRunHeader):
            self.mc_run_headers.append(o.parse())
            o = next(self)

        self.corsika_inputcards = []
        while isinstance(o, iact.InputCard):
            self.corsika_inputcards.append(o.parse())
            o = next(self)

        expected_structure = [
            CameraSettings,
            CameraOrganization,
            PixelSettings,
            DisabledPixels,
            CameraSoftwareSettings,
            DriveSettings,
            PointingCorrection,
        ]

        # in some files, the telescope description block
        # strangely contains already MCRunHeader, MCEvent, MCShower
        # we save them here for later use
        self.init_mc_showers = []
        self.init_mc_events = []

        self.telescope_descriptions = defaultdict(dict)
        first = True
        for _ in range(self.n_telescopes):
            for eventio_type in expected_structure:
                if not first:
                    o = next(self)
                first = False

                while not isinstance(o, eventio_type):
                    if isinstance(o, MCRunHeader):
                        self.mc_run_headers.append(o.parse())
                        msg = 'Unexpectd MCRunHeader in telescope description block'

                    elif isinstance(o, MCShower):
                        self.init_mc_showers.append(o.parse())
                        msg = 'Unexpectd MCShower in telescope description block'

                    elif isinstance(o, MCEvent):
                        self.init_mc_events.append(
                            (o.header.id, o.parse(), self.init_mc_showers[-1])
                        )
                        msg = 'Unexpectd MCEvent in telescope description block'
                    else:
                        msg = 'Skipping unexpected object of type {}'.format(
                            o.__class__.__name__
                        )
                    warnings.warn(msg)
                    log.warning(msg)

                    o = next(self)

                key = camel_to_snake(o.__class__.__name__)
                self.telescope_descriptions[o.telescope_id][key] = o.parse()

        self._first_event_byte = self.tell()

    def __iter__(self):
        return self.iter_array_events()

    def iter_mc_events(self):
        self._next_header_pos = self._first_event_byte

        for event_id, event, shower in self.init_mc_events:
            yield {'event_id': event_id, 'event': event, 'shower': shower}

        current_mc_shower = None
        if len(self.init_mc_showers) > 0:
            current_mc_shower = self.init_mc_showers[-1]

        o = next(self)

        while True:
            if isinstance(o, MCShower):
                current_mc_shower = o.parse()

            elif isinstance(o, MCEvent):
                yield {
                    'event_id': o.header.id,
                    'mc_shower': current_mc_shower,
                    'mc_event': o.parse()
                }

            elif isinstance(o, Histograms):
                self.histograms = o.parse()
                break

            o = next(self)

    def iter_array_events(self):
        self._next_header_pos = self._first_event_byte

        current_mc_shower = None
        current_mc_event = None
        current_mc_event_id = None

        # check if some showers or events were already read in __init__
        if len(self.init_mc_showers) > 0:
            current_mc_shower = self.init_mc_showers[-1]

        if len(self.init_mc_events) > 0:
            current_mc_event_id, current_mc_event, _ = self.init_mc_events[-1]

        current_photoelectron_sum = None
        current_photoelectrons = {}
        camera_monitorings = defaultdict(dict)
        laser_calibrations = defaultdict(dict)

        try:
            o = next(self)
        except StopIteration:
            return

        while True:
            if isinstance(o, MCShower):
                current_mc_shower = o.parse()

            elif isinstance(o, MCEvent):
                current_mc_event = o.parse()
                current_mc_event_id = o.header.id

            elif isinstance(o, iact.TelescopeData):
                current_photoelectrons = parse_photoelectrons(o)

            elif isinstance(o, MCPhotoelectronSum):
                current_photoelectron_sum = o.parse()

            elif isinstance(o, ArrayEvent):
                array_event = parse_array_event(o, self.allowed_telescopes)

                # with allowed_telescopes set, it might happen there
                # are no telescopes left
                if self.allowed_telescopes and len(array_event['telescope_events']) == 0:
                    try:
                        o = next(self)
                    except StopIteration:
                        return
                    continue

                event_data = {
                    'event_id': current_mc_event_id,
                    'mc_shower': current_mc_shower,
                    'mc_event': current_mc_event,
                    'telescope_events': array_event['telescope_events'],
                    'tracking_positions': array_event['tracking_positions'],
                    'trigger_information': array_event['trigger_information'],
                    'photoelectron_sums': current_photoelectron_sum,
                    'photoelectrons': current_photoelectrons,
                }
                event_data['camera_monitorings'] = {
                    telescope_id: copy(camera_monitorings[telescope_id])
                    for telescope_id in array_event['telescope_events'].keys()
                }
                event_data['laser_calibrations'] = {
                    telescope_id: copy(laser_calibrations[telescope_id])
                    for telescope_id in array_event['telescope_events'].keys()
                }
                yield event_data

            elif isinstance(o, CameraMonitoring):
                camera_monitorings[o.telescope_id].update(o.parse())

            elif isinstance(o, LaserCalibration):
                laser_calibrations[o.telescope_id].update(o.parse())

            elif isinstance(o, Histograms):
                self.histograms = o.parse()
                break

            try:
                o = next(self)
            except StopIteration:
                return


def parse_array_event(array_event, allowed_telescopes=None):
    '''structure of event:
        TriggerInformation[2009]  <-- this knows how many TelescopeEvents

        TelescopeEvent[2202]
        ...
        TelescopeEvent[2208]

        TrackingPosition[2101]
        ...
        TrackingPosition[2164]

        StereoReconstruction[2015]


        In words:
            1 cent event
            n tel events
            m track events (n does not need to be == m)
            1 shower
    '''
    check_type(array_event, ArrayEvent)

    telescope_events = {}
    tracking_positions = {}

    for i, o in enumerate(array_event):
        # require first element to be a TriggerInformation
        if i == 0:
            check_type(o, TriggerInformation)
            trigger_information = o.parse()

        elif isinstance(o, TelescopeEvent):
            if allowed_telescopes is None or o.telescope_id in allowed_telescopes:
                telescope_events[o.telescope_id] = parse_telescope_event(o)

        elif isinstance(o, TrackingPosition):
            if allowed_telescopes is None or o.telescope_id in allowed_telescopes:
                tracking_positions[o.telescope_id] = o.parse()

    missing_tracking = set(telescope_events.keys()) - set(tracking_positions.keys())
    if missing_tracking:
        raise NoTrackingPositions(
            'Missing tracking positions for telescopes {}'.format(
                missing_tracking
            )
        )

    return {
        'trigger_information': trigger_information,
        'telescope_events': telescope_events,
        'tracking_positions': tracking_positions,
    }


def parse_photoelectrons(telescope_data):
    check_type(telescope_data, iact.TelescopeData)

    photo_electrons = {}
    for o in telescope_data:
        check_type(o, iact.PhotoElectrons)
        photo_electrons[o.telescope_id] = o.parse()

    return photo_electrons


def parse_telescope_event(telescope_event):
    '''Parse a telescope event'''
    check_type(telescope_event, TelescopeEvent)

    event = {'pixel_lists': {}}
    for i, o in enumerate(telescope_event):

        if i == 0:
            check_type(o, TelescopeEventHeader)
            event['header'] = o.parse()

        elif isinstance(o, ADCSamples):
            event['adc_samples'] = o.parse()

        elif isinstance(o, ADCSums):
            event['adc_sums'] = o.parse()

        elif isinstance(o, PixelTiming):
            event['pixel_timing'] = o.parse()

        elif isinstance(o, ImageParameters):
            event['image_parameters'] = o.parse()

        elif isinstance(o, PixelList):
            event['pixel_lists'][o.code] = o.parse()

    return event
